! TODO: Replace all imports in other modules with accesses to this module.
!       That will allow easy integration of other standard libraries into
!       these codes.
MODULE PRECISION
  USE ISO_FORTRAN_ENV, ONLY: RT => REAL32, IT => INT64
END MODULE PRECISION
 
